
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


def xor() -> pd.DataFrame:

    xor = {'X1': {0: 0, 1: 1, 2: 1, 3: 0, 4: 0, 5: 0, 6: 1, 7: 1, 8: 0, 9: 1, 10: 1, 11: 0, 12: 0, 13: 1, 14: 1, 15: 0, 16: 0, 17: 1, 18: 1, 19: 0}, 
    'X2': {0: 1, 1: 0, 2: 1, 3: 0, 4: 1, 5: 1, 6: 0, 7: 1, 8: 0, 9: 0, 10: 1, 11: 0, 12: 1, 13: 0, 14: 1, 15: 0, 16: 1, 17: 0, 18: 1, 19: 0}, 
    'X3': {0: 0, 1: 0, 2: 0, 3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 8: 1, 9: 1, 10: 1, 11: 1, 12: 1, 13: 1, 14: 1, 15: 1, 16: 1, 17: 1, 18: 1, 19: 1}, 
    'OR': {0: 1, 1: 1, 2: 1, 3: 0, 4: 1, 5: 1, 6: 1, 7: 1, 8: 0, 9: 1, 10: 1, 11: 0, 12: 1, 13: 1, 14: 1, 15: 0, 16: 1, 17: 1, 18: 1, 19: 0}, 
    'AND': {0: 0, 1: 0, 2: 1, 3: 0, 4: 0, 5: 0, 6: 0, 7: 1, 8: 0, 9: 0, 10: 1, 11: 0, 12: 0, 13: 0, 14: 1, 15: 0, 16: 0, 17: 0, 18: 1, 19: 0}, 
    'XOR': {0: 1, 1: 1, 2: 0, 3: 0, 4: 1, 5: 1, 6: 1, 7: 0, 8: 0, 9: 1, 10: 0, 11: 0, 12: 1, 13: 1, 14: 0, 15: 0, 16: 1, 17: 1, 18: 0, 19: 0}}
    
    return pd.DataFrame(xor)

def get_direct_positive(direct, n):
    n_direct = {}
    for key in direct:
        n_direct[key] = direct[key][1][:, :n]
    return n_direct


def get_direct_unsigned(direct, n):
    n_direct = {}
    for key in direct:
        n_direct[key] = []
        n_direct[key].append(direct[key][0][:, :n])
        n_direct[key].append(direct[key][1][:, :n])
    return n_direct


def get_indirect_unsigned(indirect):
    signed_indirect = {}
    for key in indirect:
        signed_indirect[key] = [-indirect[key], indirect[key]]
    signed_indirect['class0'] = []
    return signed_indirect


def draw_features_contribs(var_x, var_y, direct, indirect):
    """
    Parameters:
    Similar parameters to Shap chains class (var_X, var_Y).
    Direct, indirect -> from Shap chains get contributions.
    Returns:
    Plot non normalized and normalized Shapley contributions with stacked plots.
    Print non normalized Shapley table.
    """

    plt.figure(figsize=(18, 18))
    n = len(var_x)  # nb of features
    k = len(var_y)  # nb of outputs
    shap_chain = pd.DataFrame(index=var_x)  # shap_chain results

    shap_chain['direct_' + str(var_y[0])] = np.mean(np.abs(direct['class0'][1]), axis=0)

    for i in range(1, k):
        shap_chain['direct_' + str(var_y[i])] = np.mean(np.abs(direct['class' + str(i)][1]), axis=0)[:n]
        shap_chain['indirect_' + str(var_y[i])] = np.mean(np.abs(indirect['class' + str(i)]), axis=0)

    shap_chain['total'] = shap_chain.sum(axis=1)

    print("Direct and indirect shapley contributions")
    shap_chain.sort_values(by='total')[list(shap_chain)[:-1]].plot.barh(figsize=(8, 8), stacked=True)
    plt.legend(loc=4)
    #plt.show()

    # sum shapley contributions for each model, in order to normalize
    sum_shap = [shap_chain['direct_' + str(var_y[0])].sum()]
    for i in range(1, k):
        sum_shap.append((shap_chain['direct_' + str(var_y[i])] + shap_chain['indirect_' + str(var_y[i])]).sum())

    # normalize shap values for each model
    shap_chain_normalized = shap_chain.copy(deep=True)
    shap_chain_normalized['direct_' + str(var_y[0])] = shap_chain_normalized['direct_' + str(var_y[0])] / sum_shap[0]
    for i in range(1, k):
        shap_chain_normalized['direct_' + str(var_y[i])] = shap_chain_normalized['direct_' + str(var_y[i])] / sum_shap[
            i]
        shap_chain_normalized['indirect_' + str(var_y[i])] = shap_chain_normalized['indirect_' + str(var_y[i])] / \
                                                             sum_shap[i]

    shap_chain_normalized.drop('total', axis=1, inplace=True)
    shap_chain_normalized['total'] = shap_chain_normalized.sum(axis=1)

    # plt.figure(figsize=(18,18))
    print("Normalized direct and indirect shapley contributions")
    shap_chain_normalized.sort_values(by='total')[list(shap_chain_normalized)[:-1]].plot.barh(figsize=(8, 8),
                                                                                              stacked=True)
    plt.legend(loc=4)
    plt.xlabel("(f) Shapley chains with order = " + str(var_y))
    plt.savefig('normalized.png', bbox_inches="tight", dpi=1000)

    #plt.show()

    shap_chain_normalized = shap_chain_normalized.sort_values(by='total')
    return shap_chain, shap_chain_normalized




import numpy as np
import shap
import re
import pandas as pd
from itertools import combinations


class ChainContrib:

    def __init__(self, data, var_x, var_y, estimators, explainer='tree'):
        self.data = data
        self.var_X = var_x
        self.var_Y = var_y
        self.n = len(var_x)
        self.m = len(var_y)
        self.l = len(data)
        self.estimators = estimators
        self.explainer_type = explainer

    def get_direct_contrib(self):
        # direct_values = OrderedDict([])
        direct_values = {}
        data_x = self.data[self.var_X].copy(deep=True)  # test dataset
        Y = self.estimators[0].predict_proba(data_x)[:, 1]  # we use probability predictions (instead of prediction)
        # explain all the predictions in the test set
        if self.explainer_type == 'Kernel':
            explainer = shap.KernelExplainer(self.estimators[0].predict_proba, data_x)
        elif self.explainer_type == 'Deep':
            explainer = shap.DeepExplainer(self.estimators[0], data_x)
        else:
            explainer = shap.TreeExplainer(self.estimators[0])
        shap_values = explainer.shap_values(data_x)
        direct_values['class0'] = shap_values
        for i in range(1, self.m):
            data_x['Y_' + str(i - 1)] = Y.copy()
            Y = self.estimators[i].predict_proba(data_x)[:, 1]
            # explain all the predictions of the test set
            if self.explainer_type == 'Kernel':
                explainer = shap.KernelExplainer(self.estimators[i].predict_proba, data_x)
            elif self.explainer_type == 'Deep':
                explainer = shap.DeepExplainer(self.estimators[i], data_x)
            else:
                explainer = shap.TreeExplainer(self.estimators[i])
            shap_values = explainer.shap_values(data_x)
            direct_values['class' + str(i)] = shap_values
        return direct_values

    @staticmethod
    def get_class_path(nb_class):
        paths = []

        def take_second(elem):
            reversed = sorted(elem, reverse=True)
            paths.append(reversed)
            return reversed

        all_classes = [i for i in range(nb_class)]
        results = [c for r in range(1, len(all_classes) + 1) for c in set(combinations(all_classes, r))]
        sorted(results, key=take_second)
        path = []
        for s in paths:
            path.append(['class' + str(x) for x in s])
        return path

    def get_class_contrib(self, combination, output_contrib, weights):
        temp = []
        for i in range(len(combination)):
            ar = combination[i]

            w = 1
            for j in range(len(ar)):
                index = ':'
                c = int(re.search(r'\d+', str(ar[0])).group())
                try:
                    index = int(re.search(r'\d+', str(ar[j + 1])).group())
                except Exception as e:
                    e
                if index == ':':
                    w *= weights[ar[j]][:self.n]
                else:
                    w *= weights[ar[j]][self.n + index]
            temp.append(output_contrib[c] * w)
        return sum(temp)

    @staticmethod
    def get_coef_shap(shap_values):
        return [np.abs(x) / np.sum(np.abs(x)) if np.sum(np.abs(x)) != 0 else 0 for x in shap_values]

    def get_weights(self, direct_contrib):
        weights = {}
        for key in direct_contrib:
            weights[key] = np.array(self.get_coef_shap(direct_contrib[key][1]))
        return weights

    def get_output_contrib(self, direct_contrib):
        output_contrib = {}
        for key in direct_contrib:
            output_contrib[key] = direct_contrib[key][1][:, self.n:]
        return output_contrib

    @staticmethod
    def get_some_data(oc, w, class_list, i):
        out_oc = {}
        out_w = {}
        for key in class_list:
            out_oc[key] = oc[key][i]
            out_w[key] = w[key][i]
        return out_oc, out_w

    def get_indirect_contrib(self, direct_contrib=None) :

        if None is direct_contrib:
            direct_contrib = self.get_direct_contrib()
        else:
            direct_contrib = direct_contrib.copy()
        output_contrib = self.get_output_contrib(direct_contrib)
        weights = self.get_weights(direct_contrib)
        indirect_contrib = {}
        for key in output_contrib:
            temp = []
            c = int(re.search(r'\d+', key).group())
            all_paths = self.get_class_path(c)
            for i in range(self.l):
                oc, w = self.get_some_data(output_contrib, weights, list(output_contrib.keys()), i)
                temp.append(self.get_class_contrib(all_paths, oc[key], w))
            indirect_contrib[key] = np.array(temp)
        indirect_contrib['class0'] = np.array([])
        return indirect_contrib





from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()
setup(
    name='shapleychains',
    version='0.0.1',
    description='',
    url= 'https://github.com/cwayad/shapleychains.git',
    long_description=long_description,
    long_description_content_type= "text/markdown",
    author='Célia Wafa AYAD',
    author_email='celiane.ayad@gmail.com',
    packages=find_packages('src'), 
    package_dir={'': 'src'},
    install_requires=[
        'pandas',
        'numpy',
        'sklearn',
        'shap',
        'matplotlib',
    ],
    extras_requere = {
        "dev": [
            "pytest>=3.7"
        ]
    }
)